<?php

namespace Drupal\eca_content\Plugin\Action;

use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\Plugin\DataType\EntityAdapter;
use Drupal\Core\Session\AccountInterface;
use Drupal\eca\Plugin\Action\ActionBase;
use Drupal\eca\Service\Conditions;
use Drupal\eca\TypedData\PropertyPathTrait;
use Drupal\user\Entity\User;

/**
 * Replaces Drupal\Core\Field\FieldUpdateActionBase.
 *
 * We need to replace the core base class because within the ECA context
 * entities should not be saved after modifying a field value.
 *
 * The replacement is achieved with PHP's class_alias(), see eca_content.module
 */
abstract class FieldUpdateActionBase extends ActionBase {

  use PropertyPathTrait;

  /**
   * Gets an array of values to be set.
   *
   * @return array
   *   Array of values with field names as keys.
   */
  abstract protected function getFieldsToUpdate();

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'load_entity_into_token' => TRUE,
      'load_current_user_into_token' => TRUE,
      'save_entity' => FALSE,
      'strip_tags' => FALSE,
      'trim' => FALSE,
    ] + parent::defaultConfiguration();
  }

  /**
   * Helper function to save the entity only outside ECA context or when
   * requested explicitly.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity which might have to be saved.
   */
  protected function save(ContentEntityInterface $entity): void {
    if (empty($entity->eca_context) || ($this->configuration['save_entity'] === Conditions::OPTION_YES)) {
      $entity->save();
    }
  }

  /**
   * {@inheritdoc}
   */
  public function execute($entity = NULL) {
    if (!($entity instanceof EntityInterface)) {
      return;
    }

    if ($this->configuration['load_entity_into_token'] === Conditions::OPTION_YES) {
      $this->tokenServices->addTokenData($this->tokenServices->getTokenType($entity), $entity);
    }
    if ($this->configuration['load_current_user_into_token'] === Conditions::OPTION_YES) {
      $user = User::load($this->currentUser->id());
      $this->tokenServices->addTokenData($this->tokenServices->getTokenType($user), $user);
    }
    $all_entities_to_save = [];
    $options = ['auto_append' => TRUE, 'access' => 'update'];
    foreach ($this->getFieldsToUpdate() as $field => $value) {
      if ($this->configuration['strip_tags'] === Conditions::OPTION_YES) {
        $value = preg_replace('/[\t\n\r\0\x0B]/', '', strip_tags($value));
      }
      if ($this->configuration['trim'] === Conditions::OPTION_YES) {
        $value = trim($value);
      }
      $metadata = [];
      if ($update_target = $this->getTypedProperty(EntityAdapter::createFromEntity($entity), $field, $options, $metadata)) {
        // Try to set the value. If that attempt fails, then it would throw an
        // exception, and the exception would be logged as an error.
        $update_target->setValue($value);
        if (empty($metadata['entities'])) {
          throw new \LogicException(sprintf("The provided field %s does not resolve for entities to be saved from the %s entity having ID %s.", $field, $entity->getEntityTypeId(), $entity->id()));
        }
        foreach ($metadata['entities'] as $entity_to_save) {
          if (!in_array($entity_to_save, $all_entities_to_save, TRUE)) {
            $all_entities_to_save[] = $entity_to_save;
          }
        }
      }
      else {
        // The property path does not exist, thus the provided field is not a
        // valid argument.
        throw new \InvalidArgumentException(sprintf("The provided field %s does not exist as a property path on the %s entity having ID %s.", $field, $entity->getEntityTypeId(), $entity->id()));
      }
    }
    foreach ($all_entities_to_save as $to_save) {
      $this->save($to_save);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function access($object, AccountInterface $account = NULL, $return_as_object = FALSE) {
    /** @var \Drupal\Core\Entity\EntityInterface $entity */
    $entity = $object;
    $entity_op = 'update';

    /** @var \Drupal\Core\Access\AccessResultInterface $result */
    $result = $entity->access($entity_op, $account, TRUE);

    $options = ['auto_append' => TRUE, 'access' => 'update'];
    foreach (array_keys($this->getFieldsToUpdate()) as $field) {
      $metadata = [];
      if ($this->getTypedProperty(EntityAdapter::createFromEntity($entity), $field, $options, $metadata)) {
        $result->andIf($metadata['access']);
      }
      else {
        throw new \InvalidArgumentException(sprintf("The provided field %s does not exist as a property path on the %s entity having ID %s.", $field, $entity->getEntityTypeId(), $entity->id()));
      }
    }

    return $return_as_object ? $result : $result->isAllowed();
  }

}
