<?php

namespace Drupal\eca\Entity\Objects;

use Drupal\Component\EventDispatcher\Event;
use Drupal\context_stack\Plugin\Context\GenericEntityContext;
use Drupal\eca\Plugin\Action\ActionInterface;
use Drupal\eca\Entity\Eca;
use Drupal\Core\Action\ActionInterface as CoreActionInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Plugin\Context\Context;
use Drupal\Core\Plugin\Context\ContextDefinition;
use Drupal\Core\TypedData\TypedDataInterface;
use Drupal\eca\Plugin\ObjectWithPluginInterface;

class EcaAction extends EcaObject implements ObjectWithPluginInterface {

  /**
   * @var \Drupal\Core\Action\ActionInterface
   */
  protected CoreActionInterface $plugin;

  /**
   * Event constructor.
   *
   * @param \Drupal\eca\Entity\Eca $eca
   * @param string $id
   * @param string $label
   * @param \Drupal\eca\Entity\Objects\EcaEvent $event
   * @param \Drupal\Core\Action\ActionInterface $plugin
   */
  public function __construct(Eca $eca, string $id, string $label, EcaEvent $event, CoreActionInterface $plugin) {
    parent::__construct($eca, $id, $label, $event);
    $this->plugin = $plugin;
  }

  /**
   * {@inheritdoc}
   */
  public function execute(EcaObject $predecessor, Event $event, array $context): bool {
    if (!parent::execute($predecessor, $event, $context)) {
      return FALSE;
    }

    $exception_thrown = FALSE;
    if ($this->plugin instanceof ActionInterface) {
      $this->plugin->setEvent($event);
    }
    $objects = $this->getObjects($this->plugin);
    $context_stack = $this->getContextStack('eca');
    foreach ($objects as $object) {
      $context_collection = NULL;
      try {
        $stack_context = NULL;
        if ($object instanceof EntityInterface) {
          $stack_context = GenericEntityContext::fromEntity($object);
        }
        elseif ($object instanceof TypedDataInterface) {
          $value = $object->getValue();
          if ($value instanceof EntityInterface) {
            $object = $value;
            $stack_context = GenericEntityContext::fromEntity($object);
          }
          else {
            $stack_context = new Context(ContextDefinition::create($object->getDataDefinition()->getDataType()), $object);
          }
        }
        if ($stack_context) {
          $type_parts = explode(':', $stack_context->getContextData()->getDataDefinition()->getDataType());
          $context_id = end($type_parts);
          $same_context_already_exists = FALSE;
          if ($existing_context = $context_stack->getContext($context_id)) {
            $same_context_already_exists = $existing_context->hasContextValue() && $existing_context->getContextValue() === $stack_context->getContextValue();
          }
          if ($same_context_already_exists) {
            $stack_context = NULL;
          }
          else {
            $context_collection = $context_stack->addContext($stack_context, $context_id);
          }
        }

        if (!$this->plugin->access($object)) {
          $this->logger()->warning('Access denied to %actionlabel (%actionid) from ECA %ecalabel (%ecaid) for event %event.', $context);
        }
        else {
          $this->plugin->execute($object);
        }
      }
      catch (\Exception $ex) {
        $context['%exmsg'] = $ex->getMessage();
        $context['%extrace'] = $ex->getTraceAsString();
        $this->logger()->error('Failed execution of %actionlabel (%actionid) from ECA %ecalabel (%ecaid) for event %event: %exmsg.\n\n%extrace', $context);
        $exception_thrown = TRUE;
      }
      finally {
        if ($context_collection) {
          // As the inner logic of a plugin may have added its own context
          // collection(s), they need to be popped off too, and after our
          // context collection got removed, we restore the children in the same
          // order they were added.
          $child_collections = [];
          while ($popped_collection = $context_stack->pop()) {
            if ($popped_collection === $context_collection) {
              foreach ($child_collections as $child_collection) {
                $context_stack->push($child_collection);
              }
              break;
            }
            else {
              array_unshift($child_collections, $popped_collection);
            }
          }
        }
      }
    }

    return !$exception_thrown;
  }

  /**
   * Get the plugin instance.
   *
   * @return \Drupal\Core\Action\ActionInterface
   *   The plugin instance.
   */
  public function getPlugin(): CoreActionInterface {
    return $this->plugin;
  }

}
