<?php

namespace Drupal\eca;

use Drupal\Component\EventDispatcher\Event;
use Drupal\context_stack\ContextStackFactoryTrait;
use Drupal\context_stack\ContextStackTrait;
use Drupal\context_stack\Plugin\Context\GenericEntityContext;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\eca\Entity\Eca;
use Drupal\eca\Entity\Objects\EcaObject;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\eca\Entity\Objects\EcaEvent;
use Drupal\eca\Event\ContentEntityEventInterface;
use Drupal\eca\Event\ContextProviderInterface;
use Drupal\eca\Plugin\CleanupInterface;
use Drupal\eca\Plugin\ObjectWithPluginInterface;
use Drupal\eca\Token\TokenInterface;

/**
 * Executes enabled ECA config regards applying events.
 */
class Processor {

  use ContextStackTrait;
  use ContextStackFactoryTrait;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * The logger channel.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected LoggerChannelInterface $logger;

  /**
   * The Token services.
   *
   * @var \Drupal\eca\Token\TokenInterface
   */
  protected TokenInterface $tokenServices;

  /**
   * A shortened list of ECA events where execution was applied.
   *
   * This list is only used as a temporary reminder for being able to recognize
   * a possible infinite recursion.
   *
   * @var array
   */
  protected array $executionHistory = [];

  /**
   * A parameterized threshold of the maximum allowed level of recursion.
   *
   * @var int
   */
  protected int $recursionThreshold;

  /**
   * A flag indicating whether an error was already logged regards recursion.
   *
   * The flag is used to prevent log flooding, as this may quickly happen when
   * infinite recursion would happen a lot. The site owner should see at least
   * one of such an error and may (hopefully) react accordingly.
   *
   * @var bool
   */
  protected bool $recursionErrorLogged = FALSE;

  /**
   * Processor constructor.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Logger\LoggerChannelInterface $logger
   *   The logger channel.
   * @param \Drupal\eca\Token\TokenInterface $token_services
   *   The Token services.
   * @param int $recursion_threshold
   *   A parameterized threshold of the maximum allowed level of recursion.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, LoggerChannelInterface $logger, TokenInterface $token_services, $recursion_threshold) {
    $this->entityTypeManager = $entity_type_manager;
    $this->logger = $logger;
    $this->tokenServices = $token_services;
    $this->recursionThreshold = (int) $recursion_threshold;
  }

  /**
   * Main method that executes ECA config regards applying events.
   *
   * @param \Drupal\Component\EventDispatcher\Event $event
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function execute(Event $event): void {
    $context = [
      '%event' => get_class($event),
    ];
    /** @var \Drupal\eca\Entity\EcaStorage $eca_storage */
    $eca_storage = $this->entityTypeManager->getStorage('eca');
    foreach ($eca_storage->loadByEvent($event) as $eca) {
      unset($context['%eventlabel'], $context['%eventid']);
      $context['%ecalabel'] = $eca->label();
      $context['%ecaid'] = $eca->id();
      foreach ($eca->getUsedEvents() as $ecaEvent) {
        $context['%eventlabel'] = $ecaEvent->getLabel();
        $context['%eventid'] = $ecaEvent->getId();
        $this->logger->debug('Check %eventlabel (%eventid) from ECA %ecalabel (%ecaid) for event %event.', $context);
        if ($ecaEvent->applies($event)) {
          // We need to check whether this is the root of all execution calls,
          // for being able to purge the execution history once it is not
          // needed anymore.
          $is_root_execution = empty($this->executionHistory);
          // Take a look for a repetitive execution order. If we find one,
          // we see it as the beginning of infinite recursion and stop.
          if (!$is_root_execution && $this->recursionThresholdSurpassed($ecaEvent)) {
            if (!$this->recursionErrorLogged) {
              $this->logger->error('Recursion within configured ECA events detected. Please adjust your ECA configuration so that it avoids infinite loops. Affected event: %eventlabel (%eventid) from ECA %ecalabel (%ecaid).', $context);
              $this->recursionErrorLogged = TRUE;
            }
            continue;
          }

          // Temporarily keep in mind what execution was applied here. If that
          // behavior starts to repeat, we'll halt the execution pipeline to
          // prevent infinite recursion.
          $this->executionHistory[] = $ecaEvent;

          // For each execution call, we create new layer(s) of context(s).
          $context_collections = [];
          // Hold the current state of Token data (see below).
          $token_data = $this->tokenServices->getTokenData();
          // Reset the data state of the Token services. This step is necessary
          // to make sure, that variables are only valid within this scope.
          // After execution of successors, the data will be cleared once again,
          // so that any change will not affect logic outside of this scope.
          $this->tokenServices->clearTokenData();
          // Push all contexts provided by the event to their according stacks.
          if ($event instanceof ContextProviderInterface) {
            foreach ($event->getContexts() as $purpose => $stack_contexts) {
              if (!($context_stack = $this->getContextStack($purpose))) {
                continue;
              }
              $collection = $this->getContextStackFactory()->createCollection();
              foreach ($stack_contexts as $id => $stack_context) {
                $collection->addContext($stack_context, $id);
              }
              $context_collections[$purpose][] = $collection;
              $context_stack->push($collection);
            }
          }
          // Take care of the ECA context stack, which will be mostly used
          // within any configured logic.
          $context_stack = $this->getContextStack('eca');
          if ($event instanceof ContentEntityEventInterface && ($entity = $event->getEntity())) {
            // If given, push the entity that belongs to the given event as a
            // new collection layer to the context stack of ECA.
            $id = 'entity';
            if (!$context_stack->hasContext($id) || ($context_stack->getContext($id)->getContextValue() !== $entity)) {
              $context_collections['eca'][] = $context_stack->addContext(GenericEntityContext::fromEntity($entity), $id);
            }
          }
          if (empty($context_collections['eca'])) {
            // We provide at least one layer of a context collection, so that
            // after successor execution, leftovers are guaranteed to be popped
            // off the stack.
            $collection = $this->getContextStackFactory()->createCollection();
            $context_collections['eca'][] = $collection;
            $context_stack->push($collection);
          }

          // Now that we have any required context, we may execute the logic.
          $this->logger->info('Start %eventlabel (%eventid) from ECA %ecalabel (%ecaid) for event %event.', $context);
          $this->executeSuccessors($eca, $ecaEvent, $event, $context);

          // After successor execution, we pop off previously added context
          // collections, so that subsequent events are not interferred by
          // runtime changes made by the configured logic for this event.
          foreach ($context_collections as $purpose => $collections) {
            $context_stack = $this->getContextStack($purpose);
            while ($collection = array_pop($collections)) {
              // We pop items until we get to our collection instance. That way
              // we make sure, that other events will not get interferred by
              // collection leftovers, possibly coming from negligent plugins.
              while ($collection !== $context_stack->pop() && $context_stack->current());
            }
          }

          // Clear the Token data once more, and restore the state of Token data
          // for the wrapping logic (if any).
          $this->tokenServices->clearTokenData();
          foreach ($token_data as $key => $data) {
            $this->tokenServices->addTokenData($key, $data);
          }

          if ($is_root_execution) {
            // Forget what we've done here. We only take care for nested
            // triggering of events regarding possible infinite recursion.
            // By resetting the array, all root-level executions will not know
            // anything from each other.
            $this->executionHistory = [];
          }
        }
      }
    }
  }

  /**
   * @param \Drupal\eca\Entity\Eca $eca
   * @param \Drupal\eca\Entity\Objects\EcaObject $eca_object
   * @param \Drupal\Component\EventDispatcher\Event $event
   * @param array $context
   */
  protected function executeSuccessors(Eca $eca, EcaObject $eca_object, Event $event, array $context): void {
    $executedSuccessorIds = [];
    foreach ($eca->getSuccessors($eca_object, $event, $context) as $successor) {
      $context['%actionlabel'] = $successor->getLabel();
      $context['%actionid'] = $successor->getId();
      if (in_array($successor->getId(), $executedSuccessorIds, TRUE)) {
        $this->logger->debug('Prevent duplicate execution of %actionlabel (%actionid) from ECA %ecalabel (%ecaid) for event %event.', $context);
        continue;
      }
      $this->logger->info('Execute %actionlabel (%actionid) from ECA %ecalabel (%ecaid) for event %event.', $context);
      if ($successor->execute($eca_object, $event, $context)) {
        $executedSuccessorIds[] = $successor->getId();
        $this->executeSuccessors($eca, $successor, $event, $context);
      }
    }
    if ($eca_object instanceof ObjectWithPluginInterface) {
      $plugin = $eca_object->getPlugin();
      if ($plugin instanceof CleanupInterface) {
        $plugin->cleanupAfterSuccessors();
      }
    }
  }

  /**
   * Checks the ECA event object whether it surpasses the recursion threshold.
   *
   * @param \Drupal\eca\Entity\Objects\EcaEvent $ecaEvent
   *   The ECA event object to check for.
   *
   * @return bool
   *   Returns TRUE when recursion threshold was surpassed, FALSE otherwise.
   */
  protected function recursionThresholdSurpassed(EcaEvent $ecaEvent): bool {
    if (!in_array($ecaEvent, $this->executionHistory, TRUE)) {
      return FALSE;
    }
    $recursion_level = 1;
    $executed_block = [];
    $entry = end($this->executionHistory);
    while ($entry) {
      array_unshift($executed_block, $entry);
      if ($entry === $ecaEvent) {
        $block_size = count($executed_block);
        break;
      }
      $entry = prev($this->executionHistory);
    }
    while (!($recursion_level > $this->recursionThreshold)) {
      $entry = end($executed_block);
      $block_index = 0;
      while ($entry) {
        if ($entry !== prev($this->executionHistory)) {
          break 2;
        }
        $block_index++;
        if ($block_index >= $block_size) {
          $recursion_level++;
          break;
        }
        $entry = prev($executed_block);
      }
    }
    return $recursion_level > $this->recursionThreshold;
  }

}
